/* -*- Mode: C; c-file-style: "stroustrup" -*- */

/* NATools - Package with parallel corpora tools
 * Copyright (C) 2002-2006  Alberto Simes
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "ngramidx.h"

#define CACHE_SIZE 100000
GHashTable * bigram_cache;
GHashTable * trigram_cache;
GHashTable * tetragram_cache;

sqlite3* ngram_index_new(const char* filename) { 
    sqlite3 *db;
    FILE *existtest;;
    char *errmsg = NULL;
    int rc;

    existtest = fopen(filename, "r");

    if (existtest) {
	fclose(existtest);

	rc = sqlite3_open(filename, &db);
	if( rc ){
	    fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(db));
	    sqlite3_close(db);
	    exit(1);
	}

	sqlite3_exec(db, "PRAGMA temp_store = MEMORY",  NULL, NULL, NULL);
	sqlite3_exec(db, "PRAGMA cache_size = 1000000", NULL, NULL, NULL);

    } else {

	rc = sqlite3_open(filename, &db);
	if( rc ){
	    fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(db));
	    sqlite3_close(db);
	    exit(1);
	}

	sqlite3_exec(db, "PRAGMA temp_store = MEMORY",  NULL, NULL, NULL);
	sqlite3_exec(db, "PRAGMA cache_size = 1000000", NULL, NULL, NULL);

	rc = sqlite3_exec(db,
			  "CREATE TABLE bigrams (word1 INTEGER, word2 INTEGER, occs INTEGER, PRIMARY KEY (word1, word2))",
			  NULL, NULL, &errmsg);
	if (rc != SQLITE_OK) {
	    fprintf(stderr, "Error creating table: %s\n", errmsg);
	    sqlite3_free(errmsg);
	    sqlite3_close(db);
	    exit(1);
	}

	rc = sqlite3_exec(db,
			  "CREATE TABLE trigrams (word1 INTEGER, word2 INTEGER, word3 INTEGER, occs INTEGER, PRIMARY KEY (word1, word2, word3))",
			  NULL, NULL, &errmsg);
	
	if (rc != SQLITE_OK) {
	    fprintf(stderr, "Error creating table: %s\n", errmsg);
	    sqlite3_free(errmsg);
	    sqlite3_close(db);
	    exit(1);
	}

	rc = sqlite3_exec(db,
			  "CREATE TABLE tetragrams (word1 INTEGER, word2 INTEGER, word3 INTEGER, word4 INTEGER, occs INTEGER, PRIMARY KEY (word1, word2, word3, word4))",
			  NULL, NULL, &errmsg);
	
	if (rc != SQLITE_OK) {
	    fprintf(stderr, "Error creating table: %s\n", errmsg);
	    sqlite3_free(errmsg);
	    sqlite3_close(db);
	    exit(1);
	}
    }

    sqlite3_exec(db, "BEGIN", NULL, NULL, NULL);

    /* Let's initialize our beloved cache :D */
    bigram_cache    = g_hash_table_new(g_str_hash, g_str_equal); 
    trigram_cache   = g_hash_table_new(g_str_hash, g_str_equal);
    tetragram_cache = g_hash_table_new(g_str_hash, g_str_equal);

    return db;
} 



sqlite3* ngram_index_open(const char* filename) { 
    sqlite3 *db;
    FILE *existtest;
    int rc;

    existtest = fopen(filename, "r");
    if (!existtest) return NULL;
    fclose(existtest);

    rc = sqlite3_open(filename, &db);
    if( rc ){
	fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(db));
	sqlite3_close(db);
	return NULL;
    }

    sqlite3_exec(db, "PRAGMA temp_store = MEMORY",  NULL, NULL, NULL);
    sqlite3_exec(db, "PRAGMA cache_size = 1000000", NULL, NULL, NULL);
    sqlite3_exec(db, "BEGIN", NULL, NULL, NULL);

    return db;
 } 


void ngram_index_close(sqlite3* db) { 
    sqlite3_exec(db, "BEGIN", NULL, NULL, NULL);

    /* Dump our final cache */
    g_hash_table_foreach_steal(bigram_cache,    bigram_free_cache,    (gpointer) db);
    g_hash_table_foreach_steal(trigram_cache,   trigram_free_cache,   (gpointer) db);
    g_hash_table_foreach_steal(tetragram_cache, tetragram_free_cache, (gpointer) db);

    /* ... and free the birds! :) */
    /* XXX: I don't f*** know, but this s*** makes glib segfault.. */
    // g_hash_table_destroy(bigram_cache);
    // g_hash_table_destroy(trigram_cache);
    // g_hash_table_destroy(tetragram_cache);

    /* COMMIT COMMIT!! */
    sqlite3_exec(db, "END", NULL, NULL, NULL);
    sqlite3_close(db);
} 

static int set_exists(void *exists, int argc, char **argv, char **azColName) {
    *((guint32*)exists) = (guint32)g_ascii_strtoull(argv[0], NULL, 10);
    return 0;
}

void bigram_add_occurrence(sqlite3* db, guint32 w1, guint32 w2) {
    guint32 * counter;
    char *token = NULL;

    /* Use our beloved cache */
    token = g_strdup_printf("%u|%u", w1, w2);
    counter = (guint32 *) g_hash_table_lookup(bigram_cache, token);

    if(counter) 
        (*counter)++;
    else {
        counter = (guint32*) g_malloc(sizeof(guint32));
        *counter = 1;
    }
    g_hash_table_insert(bigram_cache, token, counter); 

    if(g_hash_table_size(bigram_cache) > CACHE_SIZE) {
        g_hash_table_foreach_steal(bigram_cache, bigram_free_cache, (gpointer) db);
    }
    /* END cache */
}

void trigram_add_occurrence(sqlite3* db, guint32 w1, guint32 w2, guint32 w3) {
    guint32 * counter;
    char *token = NULL;

    /* Use our beloved cache */
    token = g_strdup_printf("%u|%u|%u", w1, w2, w3);
    counter = (guint32 *) g_hash_table_lookup(trigram_cache, token);

    if(counter) 
        (*counter)++;
    else {
        counter = (guint32*) g_malloc(sizeof(guint32));
        *counter = 1;
    }
    g_hash_table_insert(trigram_cache, token, counter); 

    if(g_hash_table_size(trigram_cache) > CACHE_SIZE) {
        g_hash_table_foreach_steal(trigram_cache, trigram_free_cache, (gpointer) db);
    }
    /* END cache */
}

void tetragram_add_occurrence(sqlite3* db, guint32 w1, guint32 w2, guint32 w3, guint32 w4) {
    guint32 * counter;
    char *token = NULL;

    /* Use our beloved cache */
    token = g_strdup_printf("%u|%u|%u|%u", w1, w2, w3, w4);
    counter = (guint32 *) g_hash_table_lookup(tetragram_cache, token);

    if(counter) 
        (*counter)++;
    else {
        counter = (guint32*) g_malloc(sizeof(guint32));
        *counter = 1;
    }
    g_hash_table_insert(tetragram_cache, token, counter); 

    if(g_hash_table_size(tetragram_cache) > CACHE_SIZE) {
        g_hash_table_foreach_steal(tetragram_cache, tetragram_free_cache, (gpointer) db);
    }
    /* END cache */
}

gboolean bigram_free_cache(gpointer key, gpointer value, gpointer user_data) {
    int rc;
    char * query = NULL;
    char * skey   = (char *)    key;
    guint32 * nvalue = (guint32 *) value;
    char *errmsg  = NULL;
    guint32 exists = 0;
    guint32 w1, w2;
    sqlite3 *db = (sqlite3 *) user_data;

    sscanf(skey, "%u|%u", &w1, &w2);
    
    query =  g_strdup_printf("SELECT occs FROM bigrams  WHERE word1=%u AND word2=%u",
		     w1, w2);

    rc = sqlite3_exec(db, query, set_exists, &exists, &errmsg);

    g_free(query);

    if (rc != SQLITE_OK) {
	fprintf(stderr, "Error searching for bigram: %s\n", errmsg);
	sqlite3_free(errmsg);
	sqlite3_close(db);
	exit(1);
    }

    if (exists) {
	query = g_strdup_printf("UPDATE bigrams SET occs = %u WHERE word1=%u AND word2=%u",
				exists + (*nvalue), w1, w2);
    } else {
	query = g_strdup_printf("INSERT INTO bigrams VALUES(%u,%u,%u)",
				w1, w2, *nvalue);
    }

    rc = sqlite3_exec(db, query, NULL, NULL, &errmsg);

    g_free(query);

    if (rc != SQLITE_OK) {
	fprintf(stderr, "Error inserting/updating bigram: %s\n", errmsg);
	sqlite3_free(errmsg);
	sqlite3_close(db);
	exit(1);
    }

    g_free(skey);
    g_free(nvalue);
    return TRUE;
} 

gboolean trigram_free_cache(gpointer key, gpointer value, gpointer user_data) {
    int rc;
    char * query = NULL;
    char * skey   = (char *)    key;
    guint32 * nvalue = (guint32 *) value;
    char *errmsg  = NULL;
    guint32 exists = 0;
    guint32 w1, w2, w3;
    sqlite3 *db = (sqlite3 *) user_data;

    sscanf(skey, "%u|%u|%u", &w1, &w2, &w3);

    query =  g_strdup_printf("SELECT occs FROM trigrams  WHERE word1=%u AND word2=%u AND word3=%u",
		                     w1, w2, w3);

    rc = sqlite3_exec(db, query, set_exists, &exists, &errmsg);
    g_free(query);

    if (rc != SQLITE_OK) {
	fprintf(stderr, "Error searching for trigram: %s\n", errmsg);
	sqlite3_free(errmsg);
	sqlite3_close(db);
	exit(1);
    }

    if (exists) {
	query = g_strdup_printf("UPDATE trigrams SET occs = %u WHERE word1=%u AND word2=%u AND word3=%u",
				exists + (*nvalue), w1, w2, w3);
    } else {
	query = g_strdup_printf("INSERT INTO trigrams VALUES(%u,%u,%u,%u)",
				w1, w2, w3, *nvalue);
    }

    rc = sqlite3_exec(db, query, NULL, NULL, &errmsg);

    g_free(query);


    if (rc != SQLITE_OK) {
	fprintf(stderr, "Error inserting/updating trigram: %s\n", errmsg);
	sqlite3_free(errmsg);
	sqlite3_close(db);
	exit(1);
    }

    g_free(skey);
    g_free(nvalue);
    return TRUE;
} 


gboolean tetragram_free_cache(gpointer key, gpointer value, gpointer user_data) {
    int rc;
    char * query = NULL;
    char * skey   = (char *)    key;
    guint32 * nvalue = (guint32 *) value;
    char *errmsg  = NULL;
    guint32 exists = 0;
    guint32 w1, w2, w3, w4;
    sqlite3 *db = (sqlite3 *) user_data;

    sscanf(skey, "%u|%u|%u|%u", &w1, &w2, &w3, &w4);

    query =  g_strdup_printf("SELECT occs FROM tetragrams  WHERE word1=%u AND word2=%u AND word3=%u AND word4=%u",
				             w1, w2, w3, w4);

    rc = sqlite3_exec(db, query, set_exists, &exists, &errmsg);

    g_free(query);

    if (rc != SQLITE_OK) {
	fprintf(stderr, "Error searching for tetragram: %s\n", errmsg);
	sqlite3_free(errmsg);
	sqlite3_close(db);
	exit(1);
    }

    if (exists) {
	query = g_strdup_printf("UPDATE tetragrams SET occs = %u WHERE word1=%u AND word2=%u AND word3=%u AND word4=%u",
				exists + (*nvalue), w1, w2, w3, w4);
    } else {
	query = g_strdup_printf("INSERT INTO tetragrams VALUES(%u,%u,%u,%u,%u)",
				w1, w2, w3, w4, *nvalue);
    }

    rc = sqlite3_exec(db, query, NULL, NULL, &errmsg);

    g_free(query);

    if (rc != SQLITE_OK) {
	fprintf(stderr, "Error inserting/updating tetragram: %s\n", errmsg);
	sqlite3_free(errmsg);
	sqlite3_close(db);
	exit(1);
    }

    g_free(skey);
    g_free(nvalue);
    return TRUE;
} 

